#include <stdio.h>
#include <stdlib.h>
#include <gtk/gtk.h>
#include <heapgraph.h>

static GtkWidget* drawing_area;
static GdkPixmap *pixmap = NULL;

/** this is the list for values, 
 *  even indices = heap values, 
 *  odd indices = currently filtered heap value 
 */
static int* values;

/** We use wrap around so we may start anywhere in the 
 *  above array. 
 */
static int startindex = 0;

/** Current number of values. 0 <= num_values <= width. */
static int num_values = 0;
static int width = 0;
static int height = 0;
static int last_max = 0;

/* Create a new backing pixmap of the appropriate size */
static gint configure_event (GtkWidget         *widget,
                             GdkEventConfigure *event) {
    if (pixmap) {
	g_object_unref (pixmap);
	/* handle better in the future... */
	values = realloc (values, 2 * widget->allocation.width * sizeof (*values));	
    } else {
	values = malloc (2 * widget->allocation.width * sizeof (*values));
    }
    width = widget->allocation.width;
    height = widget->allocation.height;
    startindex = 0;
    num_values = 0;
    last_max = 0;
    
    pixmap = gdk_pixmap_new (widget->window,
			     widget->allocation.width,
			     widget->allocation.height,
			     -1);
    gdk_draw_rectangle (pixmap,
			widget->style->white_gc,
			TRUE,
			0, 0,
			widget->allocation.width,
			widget->allocation.height);    
    return TRUE;
}

/* Redraw the screen from the backing pixmap */
static gint expose_event (GtkWidget      *widget,
                          GdkEventExpose *event) {
    if (event->area.width == 1 || startindex == 0) {
	gdk_draw_drawable (widget->window,
			   widget->style->fg_gc[GTK_WIDGET_STATE (widget)],
			   pixmap,
			   event->area.x, event->area.y,  /* xsrc,  ysrc */
			   event->area.x, event->area.y,  /* xdest, ydest */
			   event->area.width, event->area.height);    
    } else {
	/* TODO verify that sx1, w1 and w2 are correct. 
	 *      it seems they are but I am not sure max 
	 *      diff is 1 (one) /robo.
	 */
	int sx1 = event->area.x + startindex;
	int w1 = event->area.width - sx1;
	int w2 = startindex;
	
	gdk_draw_drawable (widget->window,
			   widget->style->fg_gc[GTK_WIDGET_STATE (widget)],
			   pixmap,
			   sx1, event->area.y,            /* xsrc,  ysrc */
			   event->area.x, event->area.y,  /* xdest, ydest */
			   w1, event->area.height);    
	
	gdk_draw_drawable (widget->window,
			   widget->style->fg_gc[GTK_WIDGET_STATE (widget)],
			   pixmap,
			   0, event->area.y,            /* xsrc,  ysrc */
			   w1, event->area.y,           /* xdest, ydest */
			   w2, event->area.height); 
    }
    return FALSE;
}

GtkWidget *get_heap_graph () {
    drawing_area = gtk_drawing_area_new ();
    gtk_widget_set_size_request (GTK_WIDGET (drawing_area), 500, 100);
    g_signal_connect (G_OBJECT (drawing_area), "expose_event",
		      G_CALLBACK (expose_event), NULL);
    g_signal_connect (G_OBJECT (drawing_area),"configure_event",
		      G_CALLBACK (configure_event), NULL);    
    gtk_widget_set_events (drawing_area, GDK_EXPOSURE_MASK | 
			   GDK_LEAVE_NOTIFY_MASK);
    return drawing_area;
}

void add_heap_size_value (int current_heap_size, 
			  int current_filtered_heap_size) {
    if (num_values < width) {    
	values[2 * num_values] = current_heap_size;
	values[2 * num_values + 1] = current_filtered_heap_size;
	num_values++; 
    } else {
	values[2 * startindex] = current_heap_size;
	values[2 * startindex + 1] = current_filtered_heap_size;
	startindex++;
	startindex %= width;
    }
    update_heap_graph ();
}

static void draw_line (int x, int max) {
    int h, f;
    h = (int)((double)values[2 * x] * height / max);
    f = (int)((double)values[2 * x + 1] * height / max);
    gdk_draw_line (pixmap, drawing_area->style->white_gc,
		   x, 0, x, height - h);
    gdk_draw_line (pixmap, drawing_area->style->bg_gc[GTK_STATE_SELECTED], 
		   x, height - h, x, height - f);
    gdk_draw_line (pixmap, drawing_area->style->bg_gc[GTK_STATE_PRELIGHT], 
		   x, height - f, x, height);
}

void update_heap_graph () {
    int max = 0;
    int i, x;
    // find max value...
    for (i = startindex, x = 0; x < num_values; x++, i++, i %= width)
	max = (values[2 * i] > max ? values[2 * i] : max);
    
    /* ok, this is not the most efficent way to draw, 
     * but it gets the job done. Whole graph may have to be
     * repainted if we have a different max value...
     * Feel free to improve.. /robo
     */
    if (last_max != max) {
	for (x = 0, i = startindex; x < num_values; x++, i++, i %= width)
	    draw_line (i, max);
	gtk_widget_queue_draw_area (drawing_area, 0, 0, num_values, height);
    } else {
	i = startindex + num_values - 1;
	i %= width;
	x = num_values - 1;
	draw_line (i, max);
	if (num_values == width) {
	    /** have to move whole graph.. */
	    gtk_widget_queue_draw_area (drawing_area, 0, 0, num_values, height);	    
	} else {
	    gtk_widget_queue_draw_area (drawing_area, x, 0, 1, height);	
	}
    }
    last_max = max;
}

/* Emacs Local Variables: */
/* Emacs mode:C */
/* Emacs c-indentation-style:"gnu" */
/* Emacs c-hanging-braces-alist:((brace-list-open)(brace-entry-open)(defun-open after)(substatement-open after)(block-close . c-snug-do-while)(extern-lang-open after)) */
/* Emacs c-cleanup-list:(brace-else-brace brace-elseif-brace space-before-funcall) */
/* Emacs c-basic-offset:4 */
/* Emacs End: */
